/************************************************************************/
/*  Program to sort mSEED blocks of a file in chronological order.     	*/
/*									*/
/*	Mario Aranha   						        */
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	mario@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2003 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "qlib2.h"
#include "params.h"
#include "defs.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

#define VERSION "1.0.04 (2009.078)"

static char *file_in = NULL, *file_out = NULL;

void print_usage(char *prog) {
  fprintf (stderr, "%s version %s\n", prog, VERSION);
  fprintf (stderr, "%s [-h] [-d] [-t] [-p] file_in [file_out]\n", prog);
  fprintf (stderr, "    -h print help message\n");
  fprintf (stderr, "    -d <flag> set debug flag: CMD=1, EXEC=2, DO=4, DETAIL=8\n");
  fprintf (stderr, "    -t preserve input file modification time\n");
  fprintf (stderr, "    -p just print time info for every record and quit\n");
  fprintf (stderr, "    file_in input mSEED file\n");
  fprintf (stderr, "    file_out sorted output mSEED file\n");

  return;
}


int parse_cmdline (int argc, char **argv)
{
   /* Variables needed for getopt. */
   extern char *optarg;
   extern int optind, opterr;
   int c;

   /*	Parse command line options.					*/
   while ( (c = getopt(argc, argv, "hd:tp")) != -1)
     switch (c) {
     case '?':
     case 'h':
       print_usage(argv[0]);
       return 1;
     case 'd':
       debug_option = atoi(optarg);
       break;
     case 't':
       preserve_time = 1;
       break;
     case 'p':
       print_time_info = 1;
       break;
   }

   if(argc < 2) {
     print_usage(argv[0]);
     return 1;
   }

   /*	Skip over all options and their arguments.			*/
   argv = &(argv[optind]);
   argc -= optind;

   file_in = argv[0];

   if(argc > 1) {
     file_out = argv[1];
   }  

   return 0;
}

static void print_time(ST_INFO *st_p) {
  BLKINFO *bp;

  for(bp=st_p->first; bp != st_p->last; bp = bp->next) {
     (void)fprintf(info, "%s.%s.%s.%s ",  bp->hdr->station_id, bp->hdr->network_id,
		   bp->hdr->channel_id,  bp->hdr->location_id);
     (void)fprintf(info, "rate=%.4lf ", sps_rate(bp->hdr->sample_rate,bp->hdr->sample_rate_mult)); 
     (void)fprintf(info, "(%s to ", time_to_str(bp->hdr->begtime, date_fmt));
     (void)fprintf(info, "%s)\n", time_to_str(bp->hdr->endtime, date_fmt));
  }

  return;
}

/* determine if blocks are out of time order */
static int
out_of_order(ST_INFO *st_p) {
  BLKINFO *bp;
  double t1, t2;

  /* determine if blocks are out of order wrt time */
  for(bp=st_p->first; bp != st_p->last; bp = bp->next) {
    t1 = int_to_nepoch(bp->hdr->begtime);
    t2 = int_to_nepoch(bp->next->hdr->begtime);
  
    if(t1 > t2) {
      if (debug(DEBUG_DETAIL)) {
	(void)fprintf(info, "%s: out of order blks: ", st_p->filename);
	(void)fprintf(info, "(%s to ",
            time_to_str(bp->hdr->begtime, date_fmt));
	(void)fprintf(info, "%s), ",
            time_to_str(bp->hdr->endtime, date_fmt));
        (void)fprintf(info, "(%s to ",
	    time_to_str(bp->next->hdr->begtime, date_fmt));
        (void)fprintf(info, "%s)\n",
	    time_to_str(bp->next->hdr->endtime, date_fmt));
      } 
      return 1;
    }
  }

  return 0;
}


/* compare func for qsort */
static int
cmpblks(const void *p1, const void *p2) {
  BLKINFO* const *bp1 = p1;
  BLKINFO* const *bp2 = p2;
  double t1 = int_to_nepoch((*bp1)->hdr->begtime);
  double t2 = int_to_nepoch((*bp2)->hdr->begtime);

  if (t1 > t2)
    return 1;

  if (t1 < t2)
    return -1;

  return 0;
}


int main(int argc, char **argv) {
  ST_INFO *st_p;
  BLKINFO **blks, b, *bp;
  int i;

  info = stderr;
  file_in = NULL;
  file_out = NULL;
  preserve_time = 0;
  print_time_info = 0;

  if(i = parse_cmdline (argc, argv)) {
    exit(i);
  }

  open_stream (&st_head_in, file_in, "default");
  st_p = st_head_in.next;

  /* Determine the block size and format of input stream */
  get_stream_info (st_p);

  /* Get the first block */
  get_block_info (st_p);

  /* Read in and save all the data */
  while ( position_stream (st_p, NULL, NULL) != EOF ) {
    store_block (st_p);
  }

  if(print_time_info) {
    print_time(st_p);
    return 0;
  }

  /* sort blocks in time order if needed */
  if( outoforder = out_of_order(st_p) ) {
    if (debug(DEBUG_CMD)) {
      fprintf (info, "Blocks need sorting, file=%s\n", st_p->filename);
    }
    blks = (BLKINFO**)malloc(st_p->num_blocks*sizeof(BLKINFO*));
    i = 0;
    st_p->cur = st_p->first;
    while(st_p->cur != NULL) {
      blks[i++] = st_p->cur;
      st_p->cur = st_p->cur->next;
    }
    
    qsort((void*)blks, (size_t)st_p->num_blocks, sizeof(BLKINFO*), cmpblks);

    /* re-number the blocks in order */
    for(i=0; i < st_p->num_blocks; i++) {
      blks[i]->blk_no = i+1;
    }

    /* write output file */
    do_write_file2 (st_p, blks, file_out);
 
    free(blks);
   } else {
    if (debug(DEBUG_CMD)) {
      fprintf (info, "Input blocks are already sorted\n");
    }

#ifdef FOO
    /* swap last and first blocks */
    b = *(st_p->first);
    *(st_p->first) = *(st_p->last);
    *(st_p->last) = b;

    st_p->last->prev = st_p->first->prev;
    st_p->first->prev = NULL;

    st_p->first->next = st_p->last->next;
    st_p->last->next = NULL;

    i = st_p->first->blk_no;
    st_p->first->blk_no = st_p->last->blk_no;
    st_p->last->blk_no = i;
#endif
    
    /* write output file */
    do_write_file (st_p, NULL, file_out);
  }

  return 0;
}
